@@ -2,11 +2,11 @@
    rdesktop: A Remote Desktop Protocol client.
    Protocol services - Clipboard functions
    Copyright (C) Erik Forsberg <forsberg@cendio.se> 2003
-   Copyright (C) Matthew Chapman 2003
+   Copyright (C) Matthew Chapman <matthewc.unsw.edu.au> 2003-2008
 
-   This program is free software; you can redistribute it and/or modify
+   This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
+   the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
@@ -15,8 +15,7 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 #include "rdesktop.h"
@@ -33,6 +32,9 @@
 
 static VCHANNEL *cliprdr_channel;
 
+static uint8 *last_formats = NULL;
+static uint32 last_formats_length = 0;
+
 static void
 cliprdr_send_packet(uint16 type, uint16 status, uint8 * data, uint32 length)
 {
@@ -50,30 +52,44 @@ cliprdr_send_packet(uint16 type, uint16 status, uint8 * data, uint32 length)
 	channel_send(s, cliprdr_channel);
 }
 
+/* Helper which announces our readiness to supply clipboard data
+   in a single format (such as CF_TEXT) to the RDP side.
+   To announce more than one format at a time, use
+   cliprdr_send_native_format_announce.
+ */
 void
-cliprdr_send_text_format_announce(void)
+cliprdr_send_simple_native_format_announce(uint32 format)
 {
 	uint8 buffer[36];
 
-	buf_out_uint32(buffer, CF_TEXT);
+	DEBUG_CLIPBOARD(("cliprdr_send_simple_native_format_announce\n"));
+
+	buf_out_uint32(buffer, format);
 	memset(buffer + 4, 0, sizeof(buffer) - 4);	/* description */
-	cliprdr_send_packet(CLIPRDR_FORMAT_ANNOUNCE, CLIPRDR_REQUEST, buffer, sizeof(buffer));
+	cliprdr_send_native_format_announce(buffer, sizeof(buffer));
 }
 
+/* Announces our readiness to supply clipboard data in multiple
+   formats, each denoted by a 36-byte format descriptor of
+   [ uint32 format + 32-byte description ].
+ */
 void
-cliprdr_send_blah_format_announce(void)
+cliprdr_send_native_format_announce(uint8 * formats_data, uint32 formats_data_length)
 {
-	uint8 buffer[36];
+	DEBUG_CLIPBOARD(("cliprdr_send_native_format_announce\n"));
 
-	buf_out_uint32(buffer, CF_OEMTEXT);
-	memset(buffer + 4, 0, sizeof(buffer) - 4);	/* description */
-	cliprdr_send_packet(CLIPRDR_FORMAT_ANNOUNCE, CLIPRDR_REQUEST, buffer, sizeof(buffer));
-}
+	cliprdr_send_packet(CLIPRDR_FORMAT_ANNOUNCE, CLIPRDR_REQUEST, formats_data,
+			    formats_data_length);
 
-void
-cliprdr_send_native_format_announce(uint8 * data, uint32 length)
-{
-	cliprdr_send_packet(CLIPRDR_FORMAT_ANNOUNCE, CLIPRDR_REQUEST, data, length);
+	if (formats_data != last_formats)
+	{
+		if (last_formats)
+			xfree(last_formats);
+
+		last_formats = xmalloc(formats_data_length);
+		memcpy(last_formats, formats_data, formats_data_length);
+		last_formats_length = formats_data_length;
+	}
 }
 
 void
@@ -81,6 +97,7 @@ cliprdr_send_data_request(uint32 format)
 {
 	uint8 buffer[4];
 
+	DEBUG_CLIPBOARD(("cliprdr_send_data_request\n"));
 	buf_out_uint32(buffer, format);
 	cliprdr_send_packet(CLIPRDR_DATA_REQUEST, CLIPRDR_REQUEST, buffer, sizeof(buffer));
 }
@@ -88,6 +105,7 @@ cliprdr_send_data_request(uint32 format)
 void
 cliprdr_send_data(uint8 * data, uint32 length)
 {
+	DEBUG_CLIPBOARD(("cliprdr_send_data\n"));
 	cliprdr_send_packet(CLIPRDR_DATA_RESPONSE, CLIPRDR_RESPONSE, data, length);
 }
 
@@ -107,15 +125,21 @@ cliprdr_process(STREAM s)
 
 	if (status == CLIPRDR_ERROR)
 	{
-		if (type == CLIPRDR_FORMAT_ACK)
+		switch (type)
 		{
-			/* FIXME: We seem to get this when we send an announce while the server is
-			   still processing a paste. Try sending another announce. */
-			cliprdr_send_text_format_announce();
-			return;
+			case CLIPRDR_FORMAT_ACK:
+				/* FIXME: We seem to get this when we send an announce while the server is
+				   still processing a paste. Try sending another announce. */
+				cliprdr_send_native_format_announce(last_formats,
+								    last_formats_length);
+				break;
+			case CLIPRDR_DATA_RESPONSE:
+				ui_clip_request_failed();
+				break;
+			default:
+				DEBUG_CLIPBOARD(("CLIPRDR error (type=%d)\n", type));
 		}
 
-		DEBUG_CLIPBOARD(("CLIPRDR error (type=%d)\n", type));
 		return;
 	}
 
@@ -144,7 +168,13 @@ cliprdr_process(STREAM s)
 	}
 }
 
-BOOL
+void
+cliprdr_set_mode(const char *optarg)
+{
+	ui_clip_set_mode(optarg);
+}
+
+RD_BOOL
 cliprdr_init(void)
 {
 	cliprdr_channel =
