@@ -1,21 +1,21 @@
 /* -*- c-basic-offset: 8 -*-
    rdesktop: A Remote Desktop Protocol client.
    Protocol services - ISO layer
-   Copyright (C) Matthew Chapman 1999-2005
-   
-   This program is free software; you can redistribute it and/or modify
+   Copyright (C) Matthew Chapman <matthewc.unsw.edu.au> 1999-2008
+   Copyright 2005-2011 Peter Astrand <astrand@cendio.se> for Cendio AB
+
+   This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
+   the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
-   
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-   
+
    You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 #include "rdesktop.h"
@@ -98,6 +98,11 @@ iso_recv_msg(uint8 * code, uint8 * rdpver)
 			next_be(s, length);
 		}
 	}
+	if (length < 4)
+	{
+		error("Bad packet header\n");
+		return NULL;
+	}
 	s = tcp_recv(s, length - 4);
 	if (s == NULL)
 		return NULL;
@@ -168,15 +173,22 @@ iso_recv(uint8 * rdpver)
 }
 
 /* Establish a connection up to the ISO layer */
-BOOL
-iso_connect(char *server, char *username)
+RD_BOOL
+iso_connect(char *server, char *username, RD_BOOL reconnect)
 {
 	uint8 code = 0;
 
 	if (!tcp_connect(server))
 		return False;
 
-	iso_send_connection_request(username);
+	if (reconnect)
+	{
+		iso_send_msg(ISO_PDU_CR);
+	}
+	else
+	{
+		iso_send_connection_request(username);
+	}
 
 	if (iso_recv_msg(&code, NULL) == NULL)
 		return False;
@@ -198,3 +210,10 @@ iso_disconnect(void)
 	iso_send_msg(ISO_PDU_DR);
 	tcp_disconnect();
 }
+
+/* reset the state to support reconnecting */
+void
+iso_reset_state(void)
+{
+	tcp_reset_state();
+}
