@@ -1,11 +1,12 @@
 /* -*- c-basic-offset: 8 -*-
    rdesktop: A Remote Desktop Protocol client.
    Protocol services - RDP layer
-   Copyright (C) Matthew Chapman 1999-2005
+   Copyright (C) Matthew Chapman <matthewc.unsw.edu.au> 1999-2008
+   Copyright 2003-2011 Peter Astrand <astrand@cendio.se> for Cendio AB
 
-   This program is free software; you can redistribute it and/or modify
+   This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
+   the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
@@ -14,14 +15,16 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 #include <time.h>
+#ifndef _WIN32
 #include <errno.h>
 #include <unistd.h>
+#endif
 #include "rdesktop.h"
+#include "ssl.h"
 
 #ifdef HAVE_ICONV
 #ifdef HAVE_ICONV_H
@@ -34,34 +37,50 @@
 #endif
 
 extern uint16 g_mcs_userid;
-extern char g_username[64];
+extern char *g_username;
 extern char g_codepage[16];
-extern BOOL g_bitmap_compression;
-extern BOOL g_orders;
-extern BOOL g_encryption;
-extern BOOL g_desktop_save;
-extern BOOL g_polygon_ellipse_orders;
-extern BOOL g_use_rdp5;
+extern RD_BOOL g_bitmap_compression;
+extern RD_BOOL g_orders;
+extern RD_BOOL g_encryption;
+extern RD_BOOL g_desktop_save;
+extern RD_BOOL g_polygon_ellipse_orders;
+extern RD_BOOL g_use_rdp5;
 extern uint16 g_server_rdp_version;
 extern uint32 g_rdp5_performanceflags;
-extern int g_server_bpp;
+extern int g_server_depth;
 extern int g_width;
 extern int g_height;
-extern BOOL g_bitmap_cache;
-extern BOOL g_bitmap_cache_persist_enable;
-extern BOOL g_rdp_compression;
+extern RD_BOOL g_bitmap_cache;
+extern RD_BOOL g_bitmap_cache_persist_enable;
+extern RD_BOOL g_numlock_sync;
+extern RD_BOOL g_pending_resize;
 
 uint8 *g_next_packet;
 uint32 g_rdp_shareid;
 
 extern RDPCOMP g_mppc_dict;
 
+/* Session Directory support */
+extern RD_BOOL g_redirect;
+extern char g_redirect_server[64];
+extern char g_redirect_domain[16];
+extern char g_redirect_password[64];
+extern char *g_redirect_username;
+extern char g_redirect_cookie[128];
+extern uint32 g_redirect_flags;
+/* END Session Directory support */
+
+extern uint32 g_reconnect_logonid;
+extern char g_reconnect_random[16];
+extern RD_BOOL g_has_reconnect_random;
+extern uint8 g_client_random[SEC_RANDOM_SIZE];
+
 #if WITH_DEBUG
 static uint32 g_packetno;
 #endif
 
 #ifdef HAVE_ICONV
-static BOOL g_iconv_works = True;
+static RD_BOOL g_iconv_works = True;
 #endif
 
 /* Receive an RDP packet */
@@ -72,7 +91,7 @@ rdp_recv(uint8 * type)
 	uint16 length, pdu_type;
 	uint8 rdpver;
 
-	if ((rdp_s == NULL) || (g_next_packet >= rdp_s->end))
+	if ((rdp_s == NULL) || (g_next_packet >= rdp_s->end) || (g_next_packet == NULL))
 	{
 		rdp_s = sec_recv(&rdpver);
 		if (rdp_s == NULL)
@@ -173,8 +192,8 @@ rdp_out_unistr(STREAM s, char *string, int len)
 			size_t i = 1, o = 4;
 			if ((iconv_h = iconv_open(WINDOWS_CODEPAGE, g_codepage)) == (iconv_t) - 1)
 			{
-				warning("rdp_out_unistr: iconv_open[%s -> %s] fail %d\n",
-					g_codepage, WINDOWS_CODEPAGE, (int) iconv_h);
+				warning("rdp_out_unistr: iconv_open[%s -> %s] fail %p\n",
+					g_codepage, WINDOWS_CODEPAGE, iconv_h);
 
 				g_iconv_works = False;
 				rdp_out_unistr(s, string, len);
@@ -231,10 +250,10 @@ rdp_out_unistr(STREAM s, char *string, int len)
  * Returns str_len of string
  */
 int
-rdp_in_unistr(STREAM s, char *string, int uni_len)
+rdp_in_unistr(STREAM s, char *string, int str_size, int in_len)
 {
 #ifdef HAVE_ICONV
-	size_t ibl = uni_len, obl = uni_len;
+	size_t ibl = in_len, obl = str_size - 1;
 	char *pin = (char *) s->p, *pout = string;
 	static iconv_t iconv_h = (iconv_t) - 1;
 
@@ -244,37 +263,60 @@ rdp_in_unistr(STREAM s, char *string, int uni_len)
 		{
 			if ((iconv_h = iconv_open(g_codepage, WINDOWS_CODEPAGE)) == (iconv_t) - 1)
 			{
-				warning("rdp_in_unistr: iconv_open[%s -> %s] fail %d\n",
-					WINDOWS_CODEPAGE, g_codepage, (int) iconv_h);
+				warning("rdp_in_unistr: iconv_open[%s -> %s] fail %p\n",
+					WINDOWS_CODEPAGE, g_codepage, iconv_h);
 
 				g_iconv_works = False;
-				return rdp_in_unistr(s, string, uni_len);
+				return rdp_in_unistr(s, string, str_size, in_len);
 			}
 		}
 
 		if (iconv(iconv_h, (ICONV_CONST char **) &pin, &ibl, &pout, &obl) == (size_t) - 1)
 		{
-			iconv_close(iconv_h);
-			iconv_h = (iconv_t) - 1;
-			warning("rdp_in_unistr: iconv fail, errno %d\n", errno);
+			if (errno == E2BIG)
+			{
+				warning("server sent an unexpectedly long string, truncating\n");
+			}
+			else
+			{
+				iconv_close(iconv_h);
+				iconv_h = (iconv_t) - 1;
+				warning("rdp_in_unistr: iconv fail, errno %d\n", errno);
 
-			g_iconv_works = False;
-			return rdp_in_unistr(s, string, uni_len);
+				g_iconv_works = False;
+				return rdp_in_unistr(s, string, str_size, in_len);
+			}
 		}
+
+		/* we must update the location of the current STREAM for future reads of s->p */
+		s->p += in_len;
+
+		*pout = 0;
 		return pout - string;
 	}
 	else
 #endif
 	{
 		int i = 0;
+		int len = in_len / 2;
+		int rem = 0;
 
-		while (i < uni_len / 2)
+		if (len > str_size - 1)
+		{
+			warning("server sent an unexpectedly long string, truncating\n");
+			len = str_size - 1;
+			rem = in_len - 2 * len;
+		}
+
+		while (i < len)
 		{
 			in_uint8a(s, &string[i++], 1);
 			in_uint8s(s, 1);
 		}
 
-		return i - 1;
+		in_uint8s(s, rem);
+		string[len] = 0;
+		return len;
 	}
 }
 
@@ -297,6 +339,7 @@ rdp_send_logon_info(uint32 flags, char *domain, char *user,
 	STREAM s;
 	time_t t = time(NULL);
 	time_t tzone;
+	uint8 security_verifier[16];
 
 	if (!g_use_rdp5 || 1 == g_server_rdp_version)
 	{
@@ -394,36 +437,51 @@ rdp_send_logon_info(uint32 flags, char *domain, char *user,
 		{
 			out_uint16_le(s, 0);
 		}
-		out_uint16_le(s, 2);
-		out_uint16_le(s, len_ip + 2);	/* Length of client ip */
-		rdp_out_unistr(s, ipaddr, len_ip);
-		out_uint16_le(s, len_dll + 2);
-		rdp_out_unistr(s, "C:\\WINNT\\System32\\mstscax.dll", len_dll);
+		/* TS_EXTENDED_INFO_PACKET */
+		out_uint16_le(s, 2);	/* clientAddressFamily = AF_INET */
+		out_uint16_le(s, len_ip + 2);	/* cbClientAddress, Length of client ip */
+		rdp_out_unistr(s, ipaddr, len_ip);	/* clientAddress */
+		out_uint16_le(s, len_dll + 2);	/* cbClientDir */
+		rdp_out_unistr(s, "C:\\WINNT\\System32\\mstscax.dll", len_dll);	/* clientDir */
 
+		/* TS_TIME_ZONE_INFORMATION */
 		tzone = (mktime(gmtime(&t)) - mktime(localtime(&t))) / 60;
 		out_uint32_le(s, tzone);
-
 		rdp_out_unistr(s, "GTB, normaltid", 2 * strlen("GTB, normaltid"));
 		out_uint8s(s, 62 - 2 * strlen("GTB, normaltid"));
-
 		out_uint32_le(s, 0x0a0000);
 		out_uint32_le(s, 0x050000);
 		out_uint32_le(s, 3);
 		out_uint32_le(s, 0);
 		out_uint32_le(s, 0);
-
 		rdp_out_unistr(s, "GTB, sommartid", 2 * strlen("GTB, sommartid"));
 		out_uint8s(s, 62 - 2 * strlen("GTB, sommartid"));
-
 		out_uint32_le(s, 0x30000);
 		out_uint32_le(s, 0x050000);
 		out_uint32_le(s, 2);
 		out_uint32(s, 0);
-		out_uint32_le(s, 0xffffffc4);
-		out_uint32_le(s, 0xfffffffe);
-		out_uint32_le(s, g_rdp5_performanceflags);
-		out_uint32(s, 0);
+		out_uint32_le(s, 0xffffffc4);	/* DaylightBias */
 
+		/* Rest of TS_EXTENDED_INFO_PACKET */
+		out_uint32_le(s, 0xfffffffe);	/* clientSessionId, consider changing to 0 */
+		out_uint32_le(s, g_rdp5_performanceflags);
+
+		/* Client Auto-Reconnect */
+		if (g_has_reconnect_random)
+		{
+			out_uint16_le(s, 28);	/* cbAutoReconnectLen */
+			/* ARC_CS_PRIVATE_PACKET */
+			out_uint32_le(s, 28);	/* cbLen */
+			out_uint32_le(s, 1);	/* Version */
+			out_uint32_le(s, g_reconnect_logonid);	/* LogonId */
+			ssl_hmac_md5(g_reconnect_random, sizeof(g_reconnect_random),
+				     g_client_random, SEC_RANDOM_SIZE, security_verifier);
+			out_uint8a(s, security_verifier, sizeof(security_verifier));
+		}
+		else
+		{
+			out_uint16_le(s, 0);	/* cbAutoReconnectLen */
+		}
 
 	}
 	s_mark_end(s);
@@ -482,7 +540,38 @@ rdp_send_input(uint32 time, uint16 message_type, uint16 device_flags, uint16 par
 	rdp_send_data(s, RDP_DATA_PDU_INPUT);
 }
 
-/* Inform the server on the contents of the persistent bitmap cache */
+/* Send a client window information PDU */
+void
+rdp_send_client_window_status(int status)
+{
+	STREAM s;
+	static int current_status = 1;
+
+	if (current_status == status)
+		return;
+
+	s = rdp_init_data(12);
+
+	out_uint32_le(s, status);
+
+	switch (status)
+	{
+		case 0:	/* shut the server up */
+			break;
+
+		case 1:	/* receive data again */
+			out_uint32_le(s, 0);	/* unknown */
+			out_uint16_le(s, g_width);
+			out_uint16_le(s, g_height);
+			break;
+	}
+
+	s_mark_end(s);
+	rdp_send_data(s, RDP_DATA_PDU_CLIENT_WINDOW_STATUS);
+	current_status = status;
+}
+
+/* Send persistent bitmap cache enumeration PDU's */
 static void
 rdp_enum_bmpcache2(void)
 {
@@ -576,7 +665,7 @@ rdp_out_bitmap_caps(STREAM s)
 	out_uint16_le(s, RDP_CAPSET_BITMAP);
 	out_uint16_le(s, RDP_CAPLEN_BITMAP);
 
-	out_uint16_le(s, g_server_bpp);	/* Preferred BPP */
+	out_uint16_le(s, g_server_depth);	/* Preferred colour depth */
 	out_uint16_le(s, 1);	/* Receive 1 BPP */
 	out_uint16_le(s, 1);	/* Receive 4 BPP */
 	out_uint16_le(s, 1);	/* Receive 8 BPP */
@@ -640,7 +729,7 @@ rdp_out_bmpcache_caps(STREAM s)
 	out_uint16_le(s, RDP_CAPSET_BMPCACHE);
 	out_uint16_le(s, RDP_CAPLEN_BMPCACHE);
 
-	Bpp = (g_server_bpp + 7) / 8;
+	Bpp = (g_server_depth + 7) / 8;	/* bytes per pixel */
 	out_uint8s(s, 24);	/* unused */
 	out_uint16_le(s, 0x258);	/* entries */
 	out_uint16_le(s, 0x100 * Bpp);	/* max cell size */
@@ -712,6 +801,18 @@ rdp_out_pointer_caps(STREAM s)
 	out_uint16_le(s, 20);	/* Cache size */
 }
 
+/* Output new pointer capability set */
+static void
+rdp_out_newpointer_caps(STREAM s)
+{
+	out_uint16_le(s, RDP_CAPSET_POINTER);
+	out_uint16_le(s, RDP_CAPLEN_NEWPOINTER);
+
+	out_uint16_le(s, 1);	/* Colour pointer */
+	out_uint16_le(s, 20);	/* Cache size */
+	out_uint16_le(s, 20);	/* Cache size for new pointers */
+}
+
 /* Output share capability set */
 static void
 rdp_out_share_caps(STREAM s)
@@ -734,6 +835,15 @@ rdp_out_colcache_caps(STREAM s)
 	out_uint16(s, 0);	/* pad */
 }
 
+/* Output brush cache capability set */
+static void
+rdp_out_brushcache_caps(STREAM s)
+{
+	out_uint16_le(s, RDP_CAPSET_BRUSHCACHE);
+	out_uint16_le(s, RDP_CAPLEN_BRUSHCACHE);
+	out_uint32_le(s, 1);	/* cache type */
+}
+
 static uint8 caps_0x0d[] = {
 	0x01, 0x00, 0x00, 0x00, 0x09, 0x04, 0x00, 0x00,
 	0x04, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
@@ -780,11 +890,22 @@ rdp_send_confirm_active(void)
 	uint32 sec_flags = g_encryption ? (RDP5_FLAG | SEC_ENCRYPT) : RDP5_FLAG;
 	uint16 caplen =
 		RDP_CAPLEN_GENERAL + RDP_CAPLEN_BITMAP + RDP_CAPLEN_ORDER +
-		RDP_CAPLEN_BMPCACHE + RDP_CAPLEN_COLCACHE +
+		RDP_CAPLEN_COLCACHE +
 		RDP_CAPLEN_ACTIVATE + RDP_CAPLEN_CONTROL +
-		RDP_CAPLEN_POINTER + RDP_CAPLEN_SHARE +
-		0x58 + 0x08 + 0x08 + 0x34 /* unknown caps */  +
-		4 /* w2k fix, why? */ ;
+		RDP_CAPLEN_SHARE +
+		RDP_CAPLEN_BRUSHCACHE + 0x58 + 0x08 + 0x08 + 0x34 /* unknown caps */  +
+		4 /* w2k fix, sessionid */ ;
+
+	if (g_use_rdp5)
+	{
+		caplen += RDP_CAPLEN_BMPCACHE2;
+		caplen += RDP_CAPLEN_NEWPOINTER;
+	}
+	else
+	{
+		caplen += RDP_CAPLEN_BMPCACHE;
+		caplen += RDP_CAPLEN_POINTER;
+	}
 
 	s = sec_init(sec_flags, 6 + 14 + caplen + sizeof(RDP_SOURCE));
 
@@ -798,23 +919,32 @@ rdp_send_confirm_active(void)
 	out_uint16_le(s, caplen);
 
 	out_uint8p(s, RDP_SOURCE, sizeof(RDP_SOURCE));
-	out_uint16_le(s, 0xd);	/* num_caps */
+	out_uint16_le(s, 0xe);	/* num_caps */
 	out_uint8s(s, 2);	/* pad */
 
 	rdp_out_general_caps(s);
 	rdp_out_bitmap_caps(s);
 	rdp_out_order_caps(s);
-	g_use_rdp5 ? rdp_out_bmpcache2_caps(s) : rdp_out_bmpcache_caps(s);
+	if (g_use_rdp5)
+	{
+		rdp_out_bmpcache2_caps(s);
+		rdp_out_newpointer_caps(s);
+	}
+	else
+	{
+		rdp_out_bmpcache_caps(s);
+		rdp_out_pointer_caps(s);
+	}
 	rdp_out_colcache_caps(s);
 	rdp_out_activate_caps(s);
 	rdp_out_control_caps(s);
-	rdp_out_pointer_caps(s);
 	rdp_out_share_caps(s);
+	rdp_out_brushcache_caps(s);
 
-	rdp_out_unknown_caps(s, 0x0d, 0x58, caps_0x0d);	/* international? */
-	rdp_out_unknown_caps(s, 0x0c, 0x08, caps_0x0c);
-	rdp_out_unknown_caps(s, 0x0e, 0x08, caps_0x0e);
-	rdp_out_unknown_caps(s, 0x10, 0x34, caps_0x10);	/* glyph cache? */
+	rdp_out_unknown_caps(s, 0x0d, 0x58, caps_0x0d);	/* CAPSTYPE_INPUT */
+	rdp_out_unknown_caps(s, 0x0c, 0x08, caps_0x0c);	/* CAPSTYPE_SOUND */
+	rdp_out_unknown_caps(s, 0x0e, 0x08, caps_0x0e);	/* CAPSTYPE_FONT */
+	rdp_out_unknown_caps(s, 0x10, 0x34, caps_0x10);	/* CAPSTYPE_GLYPHCACHE */
 
 	s_mark_end(s);
 	sec_send(s, sec_flags);
@@ -837,28 +967,29 @@ rdp_process_general_caps(STREAM s)
 static void
 rdp_process_bitmap_caps(STREAM s)
 {
-	uint16 width, height, bpp;
+	uint16 width, height, depth;
 
-	in_uint16_le(s, bpp);
+	in_uint16_le(s, depth);
 	in_uint8s(s, 6);
 
 	in_uint16_le(s, width);
 	in_uint16_le(s, height);
 
-	DEBUG(("setting desktop size and bpp to: %dx%dx%d\n", width, height, bpp));
+	DEBUG(("setting desktop size and depth to: %dx%dx%d\n", width, height, depth));
 
 	/*
-	 * The server may limit bpp and change the size of the desktop (for
+	 * The server may limit depth and change the size of the desktop (for
 	 * example when shadowing another session).
 	 */
-	if (g_server_bpp != bpp)
+	if (g_server_depth != depth)
 	{
-		warning("colour depth changed from %d to %d\n", g_server_bpp, bpp);
-		g_server_bpp = bpp;
+		warning("Remote desktop does not support colour depth %d; falling back to %d\n",
+			g_server_depth, depth);
+		g_server_depth = depth;
 	}
 	if (g_width != width || g_height != height)
 	{
-		warning("screen size changed from %dx%d to %dx%d\n", g_width, g_height,
+		warning("Remote desktop changed from %dx%d to %dx%d.\n", g_width, g_height,
 			width, height);
 		g_width = width;
 		g_height = height;
@@ -867,7 +998,7 @@ rdp_process_bitmap_caps(STREAM s)
 }
 
 /* Process server capabilities */
-void
+static void
 rdp_process_server_caps(STREAM s, uint16 length)
 {
 	int n;
@@ -926,7 +1057,8 @@ process_demand_active(STREAM s)
 	rdp_recv(&type);	/* RDP_PDU_SYNCHRONIZE */
 	rdp_recv(&type);	/* RDP_CTL_COOPERATE */
 	rdp_recv(&type);	/* RDP_CTL_GRANT_CONTROL */
-	rdp_send_input(0, RDP_INPUT_SYNCHRONIZE, 0, ui_get_numlock_state(read_keyboard_state()), 0);
+	rdp_send_input(0, RDP_INPUT_SYNCHRONIZE, 0,
+		       g_numlock_sync ? ui_get_numlock_state(read_keyboard_state()) : 0, 0);
 
 	if (g_use_rdp5)
 	{
@@ -944,12 +1076,14 @@ process_demand_active(STREAM s)
 }
 
 /* Process a colour pointer PDU */
-void
-process_colour_pointer_pdu(STREAM s)
+static void
+process_colour_pointer_common(STREAM s, int bpp)
 {
-	uint16 x, y, width, height, cache_idx, masklen, datalen;
-	uint8 *mask, *data;
-	HCURSOR cursor;
+	uint16 width, height, cache_idx, masklen, datalen;
+	sint16 x, y;
+	uint8 *mask;
+	uint8 *data;
+	RD_HCURSOR cursor;
 
 	in_uint16_le(s, cache_idx);
 	in_uint16_le(s, x);
@@ -960,11 +1094,37 @@ process_colour_pointer_pdu(STREAM s)
 	in_uint16_le(s, datalen);
 	in_uint8p(s, data, datalen);
 	in_uint8p(s, mask, masklen);
-	cursor = ui_create_cursor(x, y, width, height, mask, data);
+	if ((width != 32) || (height != 32))
+	{
+		warning("process_colour_pointer_common: " "width %d height %d\n", width, height);
+	}
+	/* sometimes x or y is out of bounds */
+	x = MAX(x, 0);
+	x = MIN(x, width - 1);
+	y = MAX(y, 0);
+	y = MIN(y, height - 1);
+	cursor = ui_create_cursor(x, y, width, height, mask, data, bpp);
 	ui_set_cursor(cursor);
 	cache_put_cursor(cache_idx, cursor);
 }
 
+/* Process a colour pointer PDU */
+void
+process_colour_pointer_pdu(STREAM s)
+{
+	process_colour_pointer_common(s, 24);
+}
+
+/* Process a New Pointer PDU - these pointers have variable bit depth */
+void
+process_new_pointer_pdu(STREAM s)
+{
+	int xor_bpp;
+
+	in_uint16_le(s, xor_bpp);
+	process_colour_pointer_common(s, xor_bpp);
+}
+
 /* Process a cached pointer PDU */
 void
 process_cached_pointer_pdu(STREAM s)
@@ -981,7 +1141,7 @@ process_system_pointer_pdu(STREAM s)
 {
 	uint16 system_pointer_type;
 
-	in_uint16(s, system_pointer_type);
+	in_uint16_le(s, system_pointer_type);
 	switch (system_pointer_type)
 	{
 		case RDP_NULL_POINTER:
@@ -1024,6 +1184,10 @@ process_pointer_pdu(STREAM s)
 			process_system_pointer_pdu(s);
 			break;
 
+		case RDP_POINTER_NEW:
+			process_new_pointer_pdu(s);
+			break;
+
 		default:
 			unimpl("Pointer message 0x%x\n", message_type);
 	}
@@ -1106,7 +1270,7 @@ process_palette(STREAM s)
 {
 	COLOURENTRY *entry;
 	COLOURMAP map;
-	HCOLOURMAP hmap;
+	RD_HCOLOURMAP hmap;
 	int i;
 
 	in_uint8s(s, 2);	/* pad */
@@ -1166,6 +1330,51 @@ process_update_pdu(STREAM s)
 	ui_end_update();
 }
 
+
+/* Process a Save Session Info PDU */
+void
+process_pdu_logon(STREAM s)
+{
+	uint32 infotype;
+	in_uint32_le(s, infotype);
+	if (infotype == INFOTYPE_LOGON_EXTENDED_INF)
+	{
+		uint32 fieldspresent;
+
+		in_uint8s(s, 2);	/* Length */
+		in_uint32_le(s, fieldspresent);
+		if (fieldspresent & LOGON_EX_AUTORECONNECTCOOKIE)
+		{
+			uint32 len;
+			uint32 version;
+
+			/* TS_LOGON_INFO_FIELD */
+			in_uint8s(s, 4);	/* cbFieldData */
+
+			/* ARC_SC_PRIVATE_PACKET */
+			in_uint32_le(s, len);
+			if (len != 28)
+			{
+				warning("Invalid length in Auto-Reconnect packet\n");
+				return;
+			}
+
+			in_uint32_le(s, version);
+			if (version != 1)
+			{
+				warning("Unsupported version of Auto-Reconnect packet\n");
+				return;
+			}
+
+			in_uint32_le(s, g_reconnect_logonid);
+			in_uint8a(s, g_reconnect_random, 16);
+			g_has_reconnect_random = True;
+			DEBUG(("Saving auto-reconnect cookie, id=%u\n", g_reconnect_logonid));
+		}
+	}
+}
+
+
 /* Process a disconnect PDU */
 void
 process_disconnect_pdu(STREAM s, uint32 * ext_disc_reason)
@@ -1176,7 +1385,7 @@ process_disconnect_pdu(STREAM s, uint32 * ext_disc_reason)
 }
 
 /* Process data PDU */
-static BOOL
+static RD_BOOL
 process_data_pdu(STREAM s, uint32 * ext_disc_reason)
 {
 	uint8 data_pdu_type;
@@ -1189,10 +1398,10 @@ process_data_pdu(STREAM s, uint32 * ext_disc_reason)
 	struct stream *ns = &(g_mppc_dict.ns);
 
 	in_uint8s(s, 6);	/* shareid, pad, streamid */
-	in_uint16(s, len);
+	in_uint16_le(s, len);
 	in_uint8(s, data_pdu_type);
 	in_uint8(s, ctype);
-	in_uint16(s, clen);
+	in_uint16_le(s, clen);
 	clen -= 18;
 
 	if (ctype & RDP_MPPC_COMPRESSED)
@@ -1242,11 +1451,18 @@ process_data_pdu(STREAM s, uint32 * ext_disc_reason)
 		case RDP_DATA_PDU_LOGON:
 			DEBUG(("Received Logon PDU\n"));
 			/* User logged on */
+			process_pdu_logon(s);
 			break;
 
 		case RDP_DATA_PDU_DISCONNECT:
 			process_disconnect_pdu(s, ext_disc_reason);
-			return True;
+
+			/* We used to return true and disconnect immediately here, but
+			 * Windows Vista sends a disconnect PDU with reason 0 when
+			 * reconnecting to a disconnected session, and MSTSC doesn't
+			 * drop the connection.  I think we should just save the status.
+			 */
+			break;
 
 		default:
 			unimpl("data PDU %d\n", data_pdu_type);
@@ -1254,22 +1470,131 @@ process_data_pdu(STREAM s, uint32 * ext_disc_reason)
 	return False;
 }
 
+/* Process redirect PDU from Session Directory */
+static RD_BOOL
+process_redirect_pdu(STREAM s /*, uint32 * ext_disc_reason */ )
+{
+	uint32 len;
+
+	/* these 2 bytes are unknown, seem to be zeros */
+	in_uint8s(s, 2);
+
+	/* read connection flags */
+	in_uint32_le(s, g_redirect_flags);
+
+	if (g_redirect_flags & PDU_REDIRECT_HAS_IP)
+	{
+		/* read length of ip string */
+		in_uint32_le(s, len);
+
+		/* read ip string */
+		rdp_in_unistr(s, g_redirect_server, sizeof(g_redirect_server), len);
+	}
+
+	if (g_redirect_flags & PDU_REDIRECT_HAS_COOKIE)
+	{
+		/* read length of cookie string */
+		in_uint32_le(s, len);
+
+		/* read cookie string (plain ASCII) */
+		if (len > sizeof(g_redirect_cookie) - 1)
+		{
+			uint32 rem = len - (sizeof(g_redirect_cookie) - 1);
+			len = sizeof(g_redirect_cookie) - 1;
+
+			warning("Unexpectedly large redirection cookie\n");
+			in_uint8a(s, g_redirect_cookie, len);
+			in_uint8s(s, rem);
+		}
+		else
+		{
+			in_uint8a(s, g_redirect_cookie, len);
+		}
+		g_redirect_cookie[len] = 0;
+	}
+
+	if (g_redirect_flags & PDU_REDIRECT_HAS_USERNAME)
+	{
+		/* read length of username string */
+		in_uint32_le(s, len);
+
+		/* read username string */
+		g_redirect_username = (char *) xmalloc(len + 1);
+		rdp_in_unistr(s, g_redirect_username, strlen(g_redirect_username), len);
+	}
+
+	if (g_redirect_flags & PDU_REDIRECT_HAS_DOMAIN)
+	{
+		/* read length of domain string */
+		in_uint32_le(s, len);
+
+		/* read domain string */
+		rdp_in_unistr(s, g_redirect_domain, sizeof(g_redirect_domain), len);
+	}
+
+	if (g_redirect_flags & PDU_REDIRECT_HAS_PASSWORD)
+	{
+		/* read length of password string */
+		in_uint32_le(s, len);
+
+		/* read password string */
+		rdp_in_unistr(s, g_redirect_password, sizeof(g_redirect_password), len);
+	}
+
+	if (g_redirect_flags & PDU_REDIRECT_DONT_STORE_USERNAME)
+	{
+		warning("PDU_REDIRECT_DONT_STORE_USERNAME set\n");
+	}
+
+	if (g_redirect_flags & PDU_REDIRECT_USE_SMARTCARD)
+	{
+		warning("PDU_REDIRECT_USE_SMARTCARD set\n");
+	}
+
+	if (g_redirect_flags & PDU_REDIRECT_INFORMATIONAL)
+	{
+		warning("PDU_REDIRECT_INFORMATIONAL set\n");
+	}
+
+	if (g_redirect_flags & PDU_REDIRECT_HAS_TARGET_FQDN)
+	{
+		warning("PDU_REDIRECT_HAS_TARGET_FQDN set\n");
+	}
+
+	if (g_redirect_flags & PDU_REDIRECT_HAS_TARGET_NETBIOS)
+	{
+		warning("PDU_REDIRECT_HAS_TARGET_NETBIOS set\n");
+	}
+
+	if (g_redirect_flags & PDU_REDIRECT_HAS_TARGET_IP_ARRAY)
+	{
+		warning("PDU_REDIRECT_HAS_TARGET_IP_ARRAY set\n");
+	}
+
+	g_redirect = True;
+
+	return True;
+}
+
 /* Process incoming packets */
-/* nevers gets out of here till app is done */
 void
-rdp_main_loop(BOOL * deactivated, uint32 * ext_disc_reason)
+rdp_main_loop(RD_BOOL * deactivated, uint32 * ext_disc_reason)
 {
 	while (rdp_loop(deactivated, ext_disc_reason))
-		;
+	{
+		if (g_pending_resize)
+		{
+			return;
+		}
+	}
 }
 
 /* used in uiports and rdp_main_loop, processes the rdp packets waiting */
-BOOL
-rdp_loop(BOOL * deactivated, uint32 * ext_disc_reason)
+RD_BOOL
+rdp_loop(RD_BOOL * deactivated, uint32 * ext_disc_reason)
 {
 	uint8 type;
-	BOOL disc = False;	/* True when a disconnect PDU was received */
-	BOOL cont = True;
+	RD_BOOL cont = True;
 	STREAM s;
 
 	while (cont)
@@ -1287,33 +1612,43 @@ rdp_loop(BOOL * deactivated, uint32 * ext_disc_reason)
 				DEBUG(("RDP_PDU_DEACTIVATE\n"));
 				*deactivated = True;
 				break;
+			case RDP_PDU_REDIRECT:
+				return process_redirect_pdu(s);
+				break;
 			case RDP_PDU_DATA:
-				disc = process_data_pdu(s, ext_disc_reason);
+				process_data_pdu(s, ext_disc_reason);
 				break;
 			case 0:
 				break;
 			default:
 				unimpl("PDU %d\n", type);
 		}
-		if (disc)
-			return False;
 		cont = g_next_packet < s->end;
 	}
 	return True;
 }
 
 /* Establish a connection up to the RDP layer */
-BOOL
+RD_BOOL
 rdp_connect(char *server, uint32 flags, char *domain, char *password,
-	    char *command, char *directory)
+	    char *command, char *directory, RD_BOOL reconnect)
 {
-	if (!sec_connect(server, g_username))
+	if (!sec_connect(server, g_username, reconnect))
 		return False;
 
 	rdp_send_logon_info(flags, domain, g_username, password, command, directory);
 	return True;
 }
 
+/* Called during redirection to reset the state to support redirection */
+void
+rdp_reset_state(void)
+{
+	g_next_packet = NULL;	/* reset the packet information */
+	g_rdp_shareid = 0;
+	sec_reset_state();
+}
+
 /* Disconnect from the RDP layer */
 void
 rdp_disconnect(void)
