@@ -2,22 +2,21 @@
    rdesktop: A Remote Desktop Protocol client.
    User interface services - X keyboard mapping
 
-   Copyright (C) Matthew Chapman 1999-2005
-   Copyright (C) Peter Astrand <peter@cendio.se> 2003
-   
-   This program is free software; you can redistribute it and/or modify
+   Copyright (C) Matthew Chapman <matthewc.unsw.edu.au> 1999-2008
+   Copyright (C) 2003-2008 Peter Astrand <astrand@cendio.se> for Cendio AB
+
+   This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; either version 2 of the License, or
+   the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
-   
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program; if not, write to the Free Software
-   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 #ifdef RDP2VNC
@@ -31,6 +30,7 @@
 #include <limits.h>
 #include <time.h>
 #include <string.h>
+#include <assert.h>
 #include "rdesktop.h"
 #include "scancodes.h"
 
@@ -46,17 +46,18 @@ extern int g_keyboard_type;
 extern int g_keyboard_subtype;
 extern int g_keyboard_functionkeys;
 extern int g_win_button_size;
-extern BOOL g_enable_compose;
-extern BOOL g_use_rdp5;
-extern BOOL g_numlock_sync;
+extern RD_BOOL g_enable_compose;
+extern RD_BOOL g_use_rdp5;
+extern RD_BOOL g_numlock_sync;
 
-static BOOL keymap_loaded;
+static RD_BOOL keymap_loaded;
 static key_translation *keymap[KEYMAP_SIZE];
+static KeySym keypress_keysyms[256];
 static int min_keycode;
 static uint16 remote_modifier_state = 0;
 static uint16 saved_remote_modifier_state = 0;
 
-static void update_modifier_state(uint8 scancode, BOOL pressed);
+static void update_modifier_state(uint8 scancode, RD_BOOL pressed);
 
 /* Free key_translation structure, including linked list */
 static void
@@ -159,7 +160,7 @@ add_sequence(char *rest, char *mapname)
 	DEBUG_KBD(("\n"));
 }
 
-BOOL
+RD_BOOL
 xkeymap_from_locale(const char *locale)
 {
 	char *str, *ptr;
@@ -275,7 +276,7 @@ xkeymap_open(const char *filename)
 	return NULL;
 }
 
-static BOOL
+static RD_BOOL
 xkeymap_read(char *mapname)
 {
 	FILE *fp;
@@ -455,7 +456,7 @@ xkeymap_init(void)
 }
 
 static void
-send_winkey(uint32 ev_time, BOOL pressed, BOOL leftkey)
+send_winkey(uint32 ev_time, RD_BOOL pressed, RD_BOOL leftkey)
 {
 	uint8 winkey;
 
@@ -503,9 +504,39 @@ reset_winkey(uint32 ev_time)
 	}
 }
 
+
+void
+set_keypress_keysym(unsigned int keycode, KeySym keysym)
+{
+	if (keycode < 8 || keycode > 255)
+		return;
+	keypress_keysyms[keycode] = keysym;
+}
+
+
+KeySym
+reset_keypress_keysym(unsigned int keycode, KeySym keysym)
+{
+	KeySym ks;
+	if (keycode < 8 || keycode > 255)
+		return keysym;
+	ks = keypress_keysyms[keycode];
+	if (ks != 0)
+	{
+		keypress_keysyms[keycode] = 0;
+	}
+	else
+	{
+		ks = keysym;
+	}
+
+	return ks;
+}
+
+
 /* Handle special key combinations */
-BOOL
-handle_special_keys(uint32 keysym, unsigned int state, uint32 ev_time, BOOL pressed)
+RD_BOOL
+handle_special_keys(uint32 keysym, unsigned int state, uint32 ev_time, RD_BOOL pressed)
 {
 	switch (keysym)
 	{
@@ -679,9 +710,31 @@ xkeymap_translate_key(uint32 keysym, unsigned int keycode, unsigned int state)
 	return tr;
 }
 
+static RD_BOOL
+is_modifier(uint8 scancode)
+{
+	switch (scancode)
+	{
+		case SCANCODE_CHAR_LSHIFT:
+		case SCANCODE_CHAR_RSHIFT:
+		case SCANCODE_CHAR_LCTRL:
+		case SCANCODE_CHAR_RCTRL:
+		case SCANCODE_CHAR_LALT:
+		case SCANCODE_CHAR_RALT:
+		case SCANCODE_CHAR_LWIN:
+		case SCANCODE_CHAR_RWIN:
+		case SCANCODE_CHAR_NUMLOCK:
+			return True;
+		default:
+			break;
+	}
+	return False;
+}
+
+
 void
 xkeymap_send_keys(uint32 keysym, unsigned int keycode, unsigned int state, uint32 ev_time,
-		  BOOL pressed, uint8 nesting)
+		  RD_BOOL pressed, uint8 nesting)
 {
 	key_translation tr, *ptr;
 	tr = xkeymap_translate_key(keysym, keycode, state);
@@ -692,17 +745,10 @@ xkeymap_send_keys(uint32 keysym, unsigned int keycode, unsigned int state, uint3
 		if (tr.scancode == 0)
 			return;
 
-		if (pressed)
-		{
-			save_remote_modifiers(tr.scancode);
-			ensure_remote_modifiers(ev_time, tr);
-			rdp_send_scancode(ev_time, RDP_KEYPRESS, tr.scancode);
-			restore_remote_modifiers(ev_time, tr.scancode);
-		}
-		else
-		{
-			rdp_send_scancode(ev_time, RDP_KEYRELEASE, tr.scancode);
-		}
+		save_remote_modifiers(tr.scancode);
+		ensure_remote_modifiers(ev_time, tr);
+		rdp_send_scancode(ev_time, pressed ? RDP_KEYPRESS : RDP_KEYRELEASE, tr.scancode);
+		restore_remote_modifiers(ev_time, tr.scancode);
 		return;
 	}
 
@@ -762,27 +808,6 @@ get_ksname(uint32 keysym)
 	return ksname;
 }
 
-static BOOL
-is_modifier(uint8 scancode)
-{
-	switch (scancode)
-	{
-		case SCANCODE_CHAR_LSHIFT:
-		case SCANCODE_CHAR_RSHIFT:
-		case SCANCODE_CHAR_LCTRL:
-		case SCANCODE_CHAR_RCTRL:
-		case SCANCODE_CHAR_LALT:
-		case SCANCODE_CHAR_RALT:
-		case SCANCODE_CHAR_LWIN:
-		case SCANCODE_CHAR_RWIN:
-		case SCANCODE_CHAR_NUMLOCK:
-			return True;
-		default:
-			break;
-	}
-	return False;
-}
-
 void
 save_remote_modifiers(uint8 scancode)
 {
@@ -795,7 +820,7 @@ save_remote_modifiers(uint8 scancode)
 void
 restore_remote_modifiers(uint32 ev_time, uint8 scancode)
 {
-	key_translation dummy;
+	key_translation dummy = { };
 
 	if (is_modifier(scancode))
 		return;
@@ -956,7 +981,7 @@ reset_modifier_keys()
 
 
 static void
-update_modifier_state(uint8 scancode, BOOL pressed)
+update_modifier_state(uint8 scancode, RD_BOOL pressed)
 {
 #ifdef WITH_DEBUG_KBD
 	uint16 old_modifier_state;
@@ -995,7 +1020,7 @@ update_modifier_state(uint8 scancode, BOOL pressed)
 			   modifier state only on Keypress */
 			if (pressed && !g_numlock_sync)
 			{
-				BOOL newNumLockState;
+				RD_BOOL newNumLockState;
 				newNumLockState =
 					(MASK_HAS_BITS
 					 (remote_modifier_state, MapNumLockMask) == False);
